/*
 * gfloatbg.c
 * author: Javier Kohen
 * purpose: This program slowly changes the color of GNOME's desktop.
 *
 * This program relies on gconf to do most of the work. Also another
 * application such as Nautilus should be managing the desktop in order
 * for the changes to be applied.
 * Based on floatbg 1.0-24 from Debian.
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include <glib.h>
#include <glib/gstdio.h>
#include <gconf/gconf.h>
#include <gconf/gconf-client.h>

#define MaxColor 	0xff
/* Multiple of 360, so the phase doesn't jump when ticks wraps around.
 * FIXME this probably should be a multiple of 360 / sf_in_hf
 */
#define MaxTime  	(0x7fff-7)
#define pi		3.1415926535
#define default_sf_in_hf	.25
#define default_s_mid	.375
#define default_s_var	.125
#define default_v_val	.87
#define default_l_time	10.00
#define PRIMARY_BG_COLOR_KEY "/desktop/gnome/background/primary_color"

char *program_name;
GConfClient* gconf;
GConfValue* value;

struct hsv { float h,s,v; };
struct rgb { float r,g,b; };

void hsv2rgb(), ticks2hsv();
float sf_in_hf = default_sf_in_hf;
float s_mid = default_s_mid;
float s_var = default_s_var;
float v_val = default_v_val;
float l_time = default_l_time;

unsigned long GetMutableColor();
void SetColor();
double atof();

char str[100];

static
void usage(void)
{
    fprintf(stderr, "usage: %s [options]\n", program_name);
    fprintf(stderr, "  where options are:\n");
    fprintf(stderr, "  -help\n");
    fprintf(stderr, "  -value <float>   (default %1.2f)\n", default_v_val);
    fprintf(stderr, "  -satmid <float>  (default %1.3f)\n", default_s_mid);
    fprintf(stderr, "  -satvar <float>  (default %1.3f)\n", default_s_var);
    fprintf(stderr, "  -fase <float>    (default %1.2f)\n", default_sf_in_hf);
    fprintf(stderr, "  -time <float>    (default %1.2f)\n", default_l_time);
    fprintf(stderr, "  value, (satmid-satvar) and (satmid+satvar) must be between 0 and 1\n");
    exit(1);
}

static
gchar *
get_lock_filename()
{
    /* Make sure only one instance of this application is running. */
    gchar *dirname = g_build_path("-", program_name, g_get_user_name(), NULL);
    gchar *lockpath = g_build_filename(g_get_tmp_dir(), dirname, ".lock", NULL);
    g_free(dirname);
    return lockpath;
}

static
void
lock()
{
    gchar *lockpath = get_lock_filename();
    gchar *lockdir = g_path_get_dirname(lockpath);
    if (-1 == g_mkdir_with_parents(lockdir, 0700)) {
	fprintf(stderr, "%s: cannot create lock file. Single instance check will be skipped: %s\n",
		program_name, strerror(errno));
    } else {
	    int lockfd = g_open(lockpath, O_RDWR | O_CREAT, S_IRWXU);
	    if (-1 == lockfd) {
		    fprintf(stderr, "%s: cannot open lock file. Single instance check will be skipped: %s\n",
			    program_name, strerror(errno));
	    } else {
		    char spid[20];
		    pid_t pid;
		    memset(spid, 0, sizeof spid);
		    int n = read(lockfd, spid, sizeof spid);
		    spid[sizeof spid - 1] = 0;
		    if (n > 0) {
			    if (sscanf(spid, "%8x", &pid) > 0) {
				    if (kill(pid, 0) == 0 || errno != ESRCH) {
					    fprintf(stderr, "%s: another copy may be running; exiting\n", program_name);
					    exit(0);
				    }
			    }
		    }
		    pid = getpid();
		    gint pidlen = g_snprintf(spid, sizeof spid, "%8x", pid);
		    lseek(lockfd, 0, SEEK_SET);
		    write(lockfd, spid, pidlen);
		    close(lockfd);
	    }
    }
    g_free(lockdir);
    g_free(lockpath);
}

static
void
unlock()
{
    gchar *lockpath = get_lock_filename();
    g_unlink(lockpath);
    g_free(lockpath);
}

static
void clear_flag_exit(int signum) {
    if (value)
	    gconf_value_free(value);
    g_object_unref(gconf);
    unlock();
    exit(0);
}

int main(argc, argv)
    int argc;
    char **argv;
{
    register int ticks;
    int i;
    struct hsv hsv;
    struct rgb rgb;
    extern char *optarg;
    struct timeval looptime, tmptime;

    program_name = g_path_get_basename(argv[0]);
    for (i = 1; i < argc; i++) {
	if (!strcmp("-help", argv[i])) {
	    usage();
	}
	if (!strcmp("-value", argv[i])) {
	    if (++i>=argc) usage();
	    v_val = atof(argv[i]);
	    continue;
	}
	if (!strcmp("-satmid", argv[i])) {
	    if (++i>=argc) usage();
	    s_mid = atof(argv[i]);
	    continue;
	}
	if (!strcmp("-satvar", argv[i])) {
	    if (++i>=argc) usage();
	    s_var = atof(argv[i]);
	    continue;
	}
	if (!strcmp("-fase", argv[i])) {
	    if (++i>=argc) usage();
	    sf_in_hf = atof(argv[i]);
	    continue;
	}
	if (!strcmp("-time", argv[i])) {
	    if (++i>=argc) usage();
	    l_time = atof(argv[i]);
	    continue;
	}
	usage();
    }

    if ( (v_val < 0) || (v_val > 1) ) usage();
    if ( ((s_mid - s_var) < 0) || ((s_mid + s_var) > 1) ) usage();

    lock();

    gconf = gconf_client_get_default();
    if (!gconf) {
	fprintf(stderr, "%s:  unable to get default gconf client\n",
		program_name);
	exit(1);
    }

    /* if we can't create the signal handler, there's not much we can */
    /* do, so throw away the return code here */
    signal(SIGHUP, (void *)clear_flag_exit);
    signal(SIGTERM, (void *)clear_flag_exit);
    signal(SIGINT, (void *)clear_flag_exit);
    signal(SIGKILL, (void *)clear_flag_exit);

    value = gconf_value_new(GCONF_VALUE_STRING);
    if (!value) {
	fprintf(stderr, "%s:  unable to creat a gconf value\n",
		program_name);
	exit(1);
    }

    srandom( (int)time(NULL) );
    ticks = (int)random() % MaxTime;

    /* Set time values for select - ie. how long to wait between loops */
    looptime.tv_sec = 0;
    looptime.tv_usec = 0;
    if (l_time >= 1) {
      looptime.tv_sec = l_time;
      l_time = l_time - looptime.tv_sec;
    }
    if ((l_time < 1) && (l_time > 0)) {
      looptime.tv_usec = 1000000*l_time;
    }

    /* The Linux kernel time share clock ticks at 100Hz on i386, */
    /* so this loop will not run more than 100 times a second. */
    /* On Alpha this is reported to be 1000Hz. */

    while ( 1 ) {
       	ticks2hsv(ticks, &hsv);
	hsv2rgb(&hsv, &rgb);
	SetColor(&rgb, gconf, value);
	ticks = (ticks + 1) % MaxTime;
	/* select seems to alter the timeval struct, so we need */
	/* an extra variable */
	tmptime = looptime;
	select(0, NULL, NULL, NULL, &tmptime);
    }
}

void SetColor (rgb, gconf, value)
    struct rgb* rgb;
    GConfClient* gconf;
    GConfValue* value;
{
    unsigned int red = (int) MaxColor * rgb->r;
    unsigned int green = (int) MaxColor * rgb->g;
    unsigned int blue = (int) MaxColor * rgb->b;

    sprintf(str,"#%.2x%.2x%.2x",red,green,blue);
    gconf_value_set_string(value, str);
    gconf_client_set(gconf, PRIMARY_BG_COLOR_KEY, value, NULL);
}

void ticks2hsv (ticks, hsv)
    int ticks;
    struct hsv *hsv;
{
    float s_fase /*, sin() */ ;

    hsv->h = ticks % 360;
    s_fase = sf_in_hf * (pi / 180) * ticks;
    hsv->s = s_mid - (s_var * sin(s_fase));
    hsv->v = v_val;
}

void hsv2rgb (hsv, rgb)
    struct hsv *hsv;
    struct rgb *rgb;
{
    int i;
    float f, p, q, r, h, v;

    v = hsv->v;
    if (hsv->s == 0) {
	rgb->r = v; rgb->g = v; rgb->b = v; }
    else {
	h = hsv->h / 60;
	i = (int) h;
	f = h - i;
	p = hsv->v * (1 - hsv->s);
	q = hsv->v * (1 - (hsv->s * f));
	r = hsv->v * (1 - (hsv->s * (1 - f)));
	switch (i) {
	    case 0: rgb->r = v; rgb->g = r; rgb->b = p; break;
	    case 1: rgb->r = q; rgb->g = v; rgb->b = p; break;
	    case 2: rgb->r = p; rgb->g = v; rgb->b = r; break;
	    case 3: rgb->r = p; rgb->g = q; rgb->b = v; break;
	    case 4: rgb->r = r; rgb->g = p; rgb->b = v; break;
	    case 5: rgb->r = v; rgb->g = p; rgb->b = q; break;
	}
    }
}
